"""Jupancon, connector to several DBs that returns pandas. Magic included."""

__version__ = "0.2.3"


from .core import (change, df_to_table, list_schemas, list_tables, peek,
                   peek_schema, query, query_raw)


# snippet to check if the environment is interactive
def is_interactive():
    import __main__ as main

    return not hasattr(main, "__file__")


def load_magics():
    if is_interactive():
        from IPython import get_ipython

        from .magic import load_ipython_extension

        load_ipython_extension(get_ipython())
        print("SQL magics loaded!")
    else:
        print("Non interactive environment detected, no magics loaded")



"""
Jupancon Configuration.

This module handles the engine and SSH tunnels connecting to the
DB of choice.
"""

import os
from collections import defaultdict
from pathlib import Path

import yaml
import redshift_connector as redc
from sqlalchemy import create_engine
from sqlalchemy.engine.url import URL
from sshtunnel import HandlerSSHTunnelForwarderError, SSHTunnelForwarder


from .defaults import CONFIG_PATH, LOCALHOST, REDSHIFT_PORT

# TODO Proper logs


class JPTConfig:
    """
    Jupancon Configuration Class - Like a factory pattern for SQLAlchemy
    engines & SSH tunnels, but without the overengineering.
    """

    def _load_config_yaml(self, name=None, configfile=None):
        """Loads the config YAML file for a particular DB, if specified"""
        path = configfile if configfile else f"{Path.home()}/{CONFIG_PATH}"
        if os.path.exists(path):
            with open(path, "r") as config_file:
                yaml_config = yaml.safe_load(config_file)
                try:
                    if "default" in yaml_config.keys() and not name:
                        name = yaml_config["default"]

                    self.name = name
                    return defaultdict(bool, yaml_config[name])

                except KeyError:
                    print(f"{name} not found in {path}")

        return defaultdict(bool)

    def _env(self, name):
        """
        Helper function, get environment variable first, otherwise get it from the YAML file. 
        NOTE: Doesn't work for all variables, but that's fine 
        """
        return os.getenv(f"JPC_{name.upper()}", default=self.config[name])


    def _get_engine_tunnel(self):
        # TODO try/catch this for missing envs
        self.use_bastion = self._env("use_bastion") 
        self.host = LOCALHOST if self.use_bastion else self._env("host") 
        self.port = self.config["port"] if self.config["port"] else REDSHIFT_PORT
        self.dbtype = os.getenv("JPC_DB_TYPE", default=self.config["type"])
        self.user =  self._env("user") or ""
        self.dbname = os.getenv("JPC_DB", default=self.config["dbname"])
        if self.dbtype == "redshift":     
            url = URL.create(
                    drivername='redshift+redshift_connector', 
                    database=self.dbname, 
                    host=self.host,
                    username=self.user,
                    password=self._env('pass') or ""
            )
            args = {"sslmode": "prefer"}
            if self._env("profile"):  
                args_iam = {
                    "iam": True, 
                    "profile": self._env("profile"),
                    "cluster_identifier": self._env("cluster"),
                    "db_user": self._env("dbuser"),
                    "db_groups": self._env("dbgroups"),
                    "auto_create": True,
                }
                args.update(args_iam)

            self.engine = create_engine(url, connect_args=args)

        elif self.dbtype == "bigquery":
            self.project = os.getenv("JPC_GCP_PROJECT", default=self.config["project"])
            self.engine = create_engine(f"bigquery://{self.project}")
        elif not self.dbtype:
            raise Exception(f"{self.name} not found")
        else:
            raise NotImplementedError
        # Opens tunnel, but only if the port is free and we are using bastion
        try:
            if self.use_bastion:
                print("Using Bastion, testing SSH tunnel...")
                self.tunnel = SSHTunnelForwarder(
                    self._env("bastion_server"),
                    ssh_username=self._env("bastion_user"),
                    remote_bind_address=(self._env("bastion_host"), REDSHIFT_PORT),
                    local_bind_address=(LOCALHOST, REDSHIFT_PORT),
                )
                self.tunnel.start()
                print("SSH tunnel works!")
                self.tunnel.close()
            else:
                self.tunnel = None

        except HandlerSSHTunnelForwarderError:
            self.tunnel = None
            print(
                "Couldn't open SSH tunnel to Bastion. "
                "Assuming tunnel is already open, will not try again."
                "Restart Kernel to retry."
            )

    def change(self, name, configfile=None):
        """
        Change the DB to query against
        """
        self.config = self._load_config_yaml(name, configfile)
        self._get_engine_tunnel()


    def __init__(self, name=None, configfile=None):
        self.change(name, configfile)


    def close_tunnel(self):
        if self.tunnel:
            self.tunnel.close()

    def connect(self):
        try:
            if self.tunnel:
                self.tunnel.start()

        except HandlerSSHTunnelForwarderError:
            print(f"Tunnel to {REDSHIFT_PORT} already opened, moving on...")

        return self.engine.connect()



import re

from IPython.core.magic import Magics, cell_magic, line_magic, magics_class

from .core import query
from .defaults import REDSHIFT_CHUNKSIZE


# The class MUST call this class decorator at creation time
@magics_class
class SqlMagics(Magics):
    @line_magic
    def select(self, line):
        """
        Line magic for SQL statements that return data
        """
        return query(f"select {line}")

    @line_magic
    def SELECT(self, line):
        """
        Line magic for SQL statements that return data (SHOUTING EDITION)
        """
        return self.select(line)

    @cell_magic
    def sql(self, line, cell):
        """
        Cell magic for SQL statements that return data
        """
        # regex to split parameters line using spaces without breaking strings
        sline = re.compile(r"""((?:[^ "']|"[^"]*"|'[^']*')+)""").split(line)[1::2]
        kwargs = {}
        for keyval in sline[1:]:
            key, value = keyval.split("=")
            kwargs[key] = value
        # if there are any parameters, 1st parameter is always df
        if sline:
            # first parameter is the DataFrame rest are mapped against the string
            self.shell.user_ns[sline[0]] = query(
                cell.format_map(kwargs), REDSHIFT_CHUNKSIZE
            )
            # return the result
            return f"Done! Result in {sline[0]}"
        else:
            # if no parameters, return DataFrame as output in the cell
            return query(cell, REDSHIFT_CHUNKSIZE)


def load_ipython_extension(ipython):
    ipython.register_magics(SqlMagics)



import os
import pytest

from jupancon.config import JPTConfig


config_path = f"{os.path.abspath(os.path.dirname(__file__))}/test_configfile.yml"

def test_jptconfig_default():
    print(config_path)
    config = JPTConfig(configfile=config_path)
    assert config.name == 'totally-real-gcp'
    assert config.dbtype == 'bigquery'
    


def test_change_non_existing():
    with pytest.raises(Exception) as execinfo:
        JPTConfig(configfile=config_path).change("totally-exists")


"""
Core functional interface, querying the DB and helper functions
"""
import pandas as pd
from sqlalchemy import text

from .config import JPTConfig
from .defaults import REDSHIFT_CHUNKSIZE

jpt = JPTConfig()


def change(name, configfile=None):
    """Change database to `name`"""
    jpt.change(name)


def query_raw(query_string):
    """
    Raw SQL, careful, you can break things here
    """
    with jpt.connect() as conn:
        return conn.execute(f"{query_string};")

    jpt.close_tunnel()


def query(query_str, chunksize=REDSHIFT_CHUNKSIZE):
    """
    Query redshift.

    returns:
        pandas.DataFrame
    """
    with jpt.connect() as conn:
        ret = pd.DataFrame()
        for chunk in pd.read_sql(text(query_str), conn, chunksize=chunksize):
            ret = pd.concat([chunk, ret])

    jpt.close_tunnel()
    return ret


def list_schemas():
    """
    List all schemas

    returns:
        pandas.DataFrame
    """
    if jpt.dbtype == "redshift":
        return query(
            """select s.nspname as table_schema,
                      s.oid as schema_id,
                      u.usename as owner
               from pg_catalog.pg_namespace s
               join pg_catalog.pg_user u on u.usesysid = s.nspowner
               order by table_schema"""
        )
    if jpt.dbtype == "bigquery":
        return query("select schema_name FROM INFORMATION_SCHEMA.SCHEMATA")

    raise NotImplementedError


def list_tables(schema):
    """
    List all tables in a schema

    returns:
        pandas.DataFrame
    """
    if jpt.dbtype == "redshift":
        return query(
            f"""
            select table_name, table_type
            from information_schema.tables t
            where t.table_schema = '{schema}'
            order by t.table_name"""
        )

    if jpt.dbtype == "bigquery":
        return query(f"SELECT * FROM {schema}.INFORMATION_SCHEMA.TABLES")

    raise NotImplementedError


def df_to_table(dataframe, schema, table, chunksize=REDSHIFT_CHUNKSIZE):
    """
    Write pandas dataframe to redshift.
    """
    with jpt.connect() as conn:
        dataframe.to_sql(
            table,
            schema=schema,
            con=conn,
            index=False,
            if_exists="replace",
            chunksize=chunksize,
        )

    jpt.close_tunnel()


def peek(table, limit=3):
    """
    Look at the first few lines of your dataset
    transposed, so all table columns fit the screen
    nicely.
    """
    return query(f"select * from {table} limit {limit}").T


def peek_schema(schema, limit=3):
    """
    Explore a whole schema by doing peek(table) in every table.
    """
    for table in list_tables(schema).table_name:
        data = peek(f"{schema}.{table}", limit)
        if not data.empty:
            print(table)
            print(f"-------------{data}\n")



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['jupancon']

package_data = \
{'': ['*'], 'jupancon': ['.ipynb_checkpoints/*']}

install_requires = \
['pandas==1.4.3',
 'sshtunnel==0.4.0',
 'pyparsing==2.4.7',
 'sqlalchemy-redshift==0.8.6',
 'sqlalchemy-bigquery==1.2.1',
 'google-cloud-bigquery-storage==2.9.1',
 'pyarrow==6.0.0',
 'redshift-connector==2.0.907']

extras_require = \
{'test': ['pytest==6.2.5',
          'jupyter==1.0.0',
          'jupyterlab==3.1.14',
          'black==21.9b0',
          'flake8==3.9.2',
          'isort==5.9.3',
          'pytest==6.2.5']}

setup(name='jupancon',
      version='0.2.3',
      description='Jupancon, connector to several DBs that returns pandas. Magic included.',
      author=None,
      author_email='Juan González-Vallinas <juan@multilayer.io>',
      url=None,
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
     )
